from typing import Union

from nidhoggr.core.repository import BaseTextureRepo
from nidhoggr.core.response import ErrorResponse, TextureStatusResponse
from nidhoggr.core.texture import TextureRequest, TextureResponse, TextureUploadRequest
from nidhoggr.requests.core import RequestsRepo


class RequestsTextureRepo(BaseTextureRepo, RequestsRepo):

    def get(self, *, request: TextureRequest) -> Union[ErrorResponse, TextureResponse]:
        return self.fetch(endpoint='/texture/get', payload=request, clazz=TextureResponse)

    def upload(self, *, request: TextureUploadRequest) -> Union[ErrorResponse, TextureStatusResponse]:
        return self.fetch(endpoint='/texture/upload', payload=request, clazz=TextureStatusResponse)

    def clear(self, *, request: TextureRequest) -> Union[ErrorResponse, TextureStatusResponse]:
        return self.fetch(endpoint='/texture/clear', payload=request, clazz=TextureStatusResponse)



from typing import Dict, Union

from nidhoggr.core.repository import BaseUserRepo
from nidhoggr.core.response import StatusResponse, ErrorResponse
from nidhoggr.core.user import User, PasswordRequest, UserRequest
from nidhoggr.requests.core import RequestsRepo


class RequestsUserRepo(BaseUserRepo, RequestsRepo):

    def get_user(self, **kwargs: Dict[str, str]) -> Union[ErrorResponse, User]:
        payload = UserRequest(**kwargs)
        return self.fetch(endpoint='/user/get', payload=payload, clazz=User)

    def check_password(self, *, clean: str, uuid: str) -> Union[ErrorResponse, StatusResponse]:
        payload = PasswordRequest(uuid=uuid, password=clean)
        return self.fetch(endpoint='/user/check_password', payload=payload, clazz=StatusResponse)

    def save_user(self, *, user: User) -> Union[ErrorResponse, User]:
        return self.fetch(endpoint='/user/save', payload=user, clazz=User)



from typing import Optional

from pydantic import BaseModel


class StatusResponse(BaseModel):
    status: bool = True

    class Config:
        allow_mutation = False


class ErrorResponse(BaseModel):
    reason: str
    exception: Optional[str] = None

    class Config:
        allow_mutation = False


class TextureStatusResponse(BaseModel):
    message: str

    class Config:
        allow_mutation = False



from abc import ABCMeta
from typing import Union, TypeVar, Type

from pydantic import BaseModel, ValidationError
from requests import Session, RequestException

from nidhoggr.core.response import ErrorResponse

T = TypeVar('T', bound=BaseModel)


class RequestsRepo(metaclass=ABCMeta):
    _session: Session
    _api_url: str
    _timeout: int

    def __init__(self, *, api_url: str, bearer_token: str, timeout: int = 1):
        session = Session()
        session.headers.update({
            'Authorization': f'Bearer {bearer_token}',
            'Content-Type': 'application/json'
        })
        self._session = session
        self._api_url = api_url
        self._timeout = timeout

    def fetch(self, *, endpoint: str, payload: BaseModel, clazz: Type[T]) -> Union[ErrorResponse, T]:
        try:
            response = self._session.post(f'{self._api_url}{endpoint}', json=payload.dict(), timeout=self._timeout)
        except RequestException as e:
            return ErrorResponse(reason=f"Failed to fetch {endpoint}", exception=e)

        if response.status_code != 200:
            return ErrorResponse(reason=f"Expected 200, got {response.status_code} for {endpoint}")

        result = response.json()
        if not result:
            return ErrorResponse(reason=f"Response from {endpoint} didn't contain valid JSON, but {response.text}")

        try:
            return clazz.parse_obj(result)
        except ValidationError as e:
            return ErrorResponse(reason=f"Received malformed {clazz} object from API, got {result}", exception=e)
